from hitchstory.result import ResultList
from hitchstory.story import Story
from copy import copy


class StoryList(object):
    """
    A sequence of stories ready to be played in order.
    """

    def __init__(self, stories):
        for story in stories:
            assert type(story) is Story
        self._stories = stories
        self._continue_on_failure = False

    def continue_on_failure(self):
        new_story_list = copy(self)
        new_story_list._continue_on_failure = True
        return new_story_list

    def play(self):
        results = ResultList()
        if len(self._stories) > 0:
            for story in self._stories:
                result = story.play()
                results.append(result)

                if hasattr(story.engine, "_aborted") and story.engine._aborted:
                    break

                if not result.passed and not self._continue_on_failure:
                    break
        else:
            print("No stories found")
        return results

    def __len__(self):
        return len(self._stories)

    def __getitem__(self, index):
        return self._stories[index]



from strictyaml import load, Map, Str, Seq, Optional, MapPattern, Any, YAMLError
from hitchstory import exceptions, utils
from hitchstory.story import Story
from path import Path
import copy


class StoryFile(object):
    """
    YAML file containing one or more named stories, part of a collection.
    """

    def __init__(self, filename, collection):
        self._filename = Path(filename).abspath()
        self._yaml = self._filename.bytes().decode("utf8")
        self._collection = collection
        self._updated_yaml = None

        steps_schema = Seq(Str() | MapPattern(Str(), Any(), maximum_keys=1))

        story_schema = {
            Optional("steps"): steps_schema,
            Optional("about"): Str(),
            Optional("with"): Any(),
            Optional("given"): self.engine.given_definition.preconditions,
        }

        variation_schema = {
            Optional("steps"): steps_schema,
            Optional("about"): Str(),
            Optional("with"): Any(),
            Optional("given"): self.engine.given_definition.preconditions,
        }

        for info_property, info_property_schema in self.engine.info_definition.items():
            story_schema[Optional(info_property)] = info_property_schema
            variation_schema[Optional(info_property)] = info_property_schema

        story_schema[Optional("based_on")] = Str()
        story_schema[Optional("variations")] = MapPattern(
            Str(), Map(variation_schema, key_validator=utils.UnderscoredSlug())
        )

        try:
            self._parsed_yaml = load(
                self._yaml,
                Str()
                | MapPattern(
                    Str(), Map(story_schema, key_validator=utils.UnderscoredSlug())
                ),
            )
        except YAMLError as error:
            raise exceptions.StoryYAMLError(filename, str(error))

    @property
    def engine(self):
        return self._collection.engine

    def rewrite(self):
        """
        Rewrite all changes back to the file.
        """
        if self._updated_yaml is not None:
            self.path.write_text(self._updated_yaml.as_yaml())

    def update(self, story, step, kwargs):
        """
        Update a specific step in a particular story during a test run.
        """
        if self._updated_yaml is None:
            self._updated_yaml = copy.copy(self._parsed_yaml)
        if story.variation:
            if step.child_index >= 0:
                yaml_story = self._updated_yaml[story.based_on]["variations"][
                    story.child_name
                ]
                if step.arguments.single_argument:
                    yaml_story["steps"][step.child_index][step.name] = list(
                        kwargs.values()
                    )[0]
                else:
                    for key, value in kwargs.items():
                        yaml_story["steps"][step.child_index][step.name][key] = value
            else:
                yaml_story = self._updated_yaml[story.based_on]
                if step.arguments.single_argument:
                    yaml_story["steps"][step.index][step.name] = list(kwargs.values())[
                        0
                    ]
                else:
                    for key, value in kwargs.items():
                        yaml_story["steps"][step.index][step.name][key] = value
        else:
            step_to_update = self._updated_yaml[story.name]["steps"][step.index]
            if step.arguments.single_argument:
                step_to_update[step.name] = list(kwargs.values())[0]
            else:
                for key_to_update, value_to_update in kwargs.items():
                    step_to_update[step.name][key_to_update] = value_to_update

    @property
    def filename(self):
        return self._filename

    @property
    def collection(self):
        return self._collection

    @property
    def path(self):
        return Path(self._filename)

    def ordered_by_file(self):
        """
        Return all of the stories in the file in the order they
        appear in the file.
        """
        stories = []
        if self._parsed_yaml.is_mapping():
            for name, parsed_main_story in self._parsed_yaml.items():
                base_story = Story(self, str(name), parsed_main_story)
                stories.append(base_story)
                variations = []

                for variation_name, parsed_var_name in (
                    self._parsed_yaml[name].get("variations", {}).items()
                ):
                    variations.append(
                        Story(
                            self,
                            variation_name,
                            parsed_var_name,
                            variation_of=str(name),
                        )
                    )

                stories.extend(variations)
                base_story.variations = variations
        return stories



from strictyaml import Regex, ScalarValidator
from re import compile
from path import Path
import prettystack
from jinja2.environment import Environment
from jinja2 import DictLoader
from slugify import slugify
import sys


SRC_DIR = Path(__file__).realpath().dirname()

TEMPLATE_DIR = SRC_DIR.joinpath("templates")


DEFAULT_STACK_TRACE = (
    prettystack.PrettyStackTemplate()
    .to_console()
    .cut_calling_code(SRC_DIR.joinpath("story.py"))
)


PARAM_REGEX = r"^\(\((.*?)\)\)$"


YAML_Param = Regex(PARAM_REGEX)


class UnderscoredSlug(ScalarValidator):
    def validate_scalar(self, chunk):
        return slugify(chunk.contents, separator="_")


def is_parameter(text):
    """
    Is the chunk of YAML data passed to us a parameter?

    i.e. like so (( parametername ))
    """
    return isinstance(text, str) and compile(PARAM_REGEX).match(text) is not None


def parameter_name(text):
    """
    Return parameter name from parameter text.

    e.g. (( param_name )) -> "param_name"
    """
    return compile(PARAM_REGEX).match(text).group(1).strip()


def to_underscore_style(text):
    """Changes "Something like this" to "something_like_this"."""
    text = text.lower().replace(" ", "_").replace("-", "_")
    return "".join(x for x in text if x.isalpha() or x.isdigit() or x == "_")


def underscore_slugify(text):
    """Changes "Something like this" to "something_like_this"."""
    return slugify(text, separator="_")


def render_template(templates_dict, template_name, parameters):
    """
    Render a jinja2 template.
    """
    env = Environment()
    env.loader = DictLoader(templates_dict)
    return env.get_template(template_name).render(**parameters)


def current_stack_trace_data():
    """
    Build a list of tracebacks from the last stack trace
    including line numbers and filenames.

    All the data needed to build a pretty stacktrace.
    """
    tb_id = 0
    _, exception, tb = sys.exc_info()

    if exception is None:
        return None

    # Create list of tracebacks
    tracebacks = []
    while tb is not None:
        filename = tb.tb_frame.f_code.co_filename
        if filename == "<frozen importlib._bootstrap>":
            break

        tracebacks.append(
            {
                "tb_id": tb_id,
                "filename": tb.tb_frame.f_code.co_filename,
                "line": tb.tb_lineno,
                "function": tb.tb_frame.f_code.co_name,
            }
        )

        tb_id = tb_id + 1
        tb = tb.tb_next

    return {
        "tracebacks": tracebacks,
        "exception_string": str(exception),
        "docstring": exception.__doc__ if exception.__doc__ is not None else None,
        "exception_type": "{}.{}".format(
            type(exception).__module__, type(exception).__name__
        ),
    }



from hitchstory.story_collection import StoryCollection
from hitchstory.engine import validate
from hitchstory.engine import no_stacktrace_for
from hitchstory.engine import about
from hitchstory.engine import InfoDefinition
from hitchstory.engine import InfoProperty
from hitchstory.engine import GivenDefinition
from hitchstory.engine import GivenProperty
from hitchstory.engine import BaseEngine
from hitchstory.exceptions import HitchStoryException
from hitchstory.exceptions import Failure


__version__ = "0.13.2"



# -*- coding: utf-8 -*
from setuptools.command.install import install
from setuptools import find_packages
from setuptools import setup
import subprocess
import codecs
import sys
import os


def read(*parts):
    # intentionally *not* adding an encoding option to open
    # see here: https://github.com/pypa/virtualenv/issues/201#issuecomment-3145690
    return codecs.open(
        os.path.join(os.path.abspath(os.path.dirname(__file__)), *parts), "r"
    ).read()


setup(
    name="hitchstory",
    version=read("VERSION").replace("\n", ""),
    description="Type-safe, YAML-based BDD, TDD & specification by example framework for python.",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
        "Topic :: Software Development :: Libraries",
        "Operating System :: Unix",
        "Environment :: Console",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
    ],
    keywords="hitchdev framework bdd tdd declarative tests testing yaml",
    author="Colm O'Connor",
    author_email="colm.oconnor.github@gmail.com",
    url="https://hitchdev.com/hitchstory/",
    license="MIT",
    install_requires=[
        "strictyaml>=0.15.3",
        "path.py",
        "jinja2",
        "colorama",
        "pathquery>=0.2.0",
        "python-slugify>=1.2.1",
        "prettystack>=0.3.0",
        "jinja2",
    ],
    packages=find_packages(
        exclude=[
            "docs",
        ]
    ),
    package_data={},
    zip_safe=False,
    include_package_data=True,
)



"""Documentation objects for use in templates."""
from hitchstory.step_method import StepMethod
import pathlib


class DocVariation(object):
    def __init__(self, templates, variables):
        self._documentation = templates.variation(**variables)

    def documentation(self):
        return self._documentation


class DocInfoProperty(object):
    def __init__(self, templates, name, info_property):
        self._documentation = templates.info_from_name(name, info_property)

    def documentation(self):
        return self._documentation


class DocGivenProperty(object):
    def __init__(self, templates, name, given_property):
        self._documentation = templates.given_from_name(name, given_property)

    def documentation(self):
        return self._documentation


class DocGivenProperties(object):
    def __init__(self, templates, given):
        self._property_docs = {
            name: DocGivenProperty(templates, name, given[name])
            for name, given_property in given.items()
        }

    def __getattr__(self, name):
        return self._property_docs[name]

    def items(self):
        return self._property_docs.items()


class DocStep(object):
    def __init__(self, templates, step):
        step_method = StepMethod(step.step_method)
        arguments = {name: None for name in step_method.argspec.args[1:]}

        if step.arguments.is_none:
            arguments = {}
        elif step.arguments.single_argument:
            var_name = step_method.argspec.args[1:][0]
            arguments[var_name] = step.arguments.data
        else:
            if step_method.argspec.keywords:
                var_name = step_method.argspec.keywords
                arguments[var_name] = step.arguments.data
            else:
                arguments.update(step.arguments.data)

        self._documentation = templates.step_from_slug(step.slug, arguments)

    def documentation(self):
        return self._documentation


def _story_variables(story, doc_templates):
    return {
        "info": {
            name: DocInfoProperty(doc_templates, name, info_property)
            for name, info_property in story.info.items()
        },
        "slug": story.slug,
        "given": DocGivenProperties(doc_templates, story.given),
        "name": story.name,
        "about": story.about,
        "steps": [DocStep(doc_templates, step) for step in story.steps],
        "filename": pathlib.Path(story.filename),
    }


def story_template(story, doc_templates):
    variables = _story_variables(story, doc_templates)
    variables["variations"] = [
        DocVariation(doc_templates, _story_variables(variation, doc_templates))
        for variation in story.variations
    ]
    return doc_templates.story(**variables)
